package cz.kebrt.unitcheck.impl;

import gov.nasa.jpf.Error;
import gov.nasa.jpf.Property;
import gov.nasa.jpf.search.Search;
import gov.nasa.jpf.search.SearchListener;

import java.util.List;

import org.apache.log4j.Logger;

import cz.kebrt.unitcheck.api.UnitCheckListener;
import cz.kebrt.unitcheck.api.UnitChecker;
import cz.kebrt.unitcheck.api.data.CheckRunResult;
import cz.kebrt.unitcheck.api.data.Description;
import cz.kebrt.unitcheck.api.data.ExceptionFailure;
import cz.kebrt.unitcheck.api.data.PathHolder;
import cz.kebrt.unitcheck.api.data.PropertyFailure;
import cz.kebrt.unitcheck.api.data.Success;
import cz.kebrt.unitcheck.api.data.TestMethod;
import cz.kebrt.unitcheck.jpf.PathConvertor;
import cz.kebrt.unitcheck.jpf.SnapshotConvertor;
import cz.kebrt.unitcheck.jpf.StopExecutionProperty;
import cz.kebrt.unitcheck.junit.TestRunListener;
import cz.kebrt.unitcheck.junit.holder.TestRunResult;

/**
 * Implements JPF and JUnit listeners and notifies instances of {@link UnitCheckListener}
 * with summarized information. An instance of this class is created for each test checked
 * by UnitCheck.
 */
public class JPFJUnitListener implements SearchListener, TestRunListener {

    private static final Logger log = Logger.getLogger(JPFJUnitListener.class);

    private final List<UnitCheckListener> listeners;
    private final UnitChecker unitCheck;
    private final TestMethod testMethod;
    private final StopExecutionProperty haltProperty;

    /** Summarized results of checking the test. */
    private final CheckRunResult runResult;

    /** Was an error found while checking the test. */
    private boolean errorReported = false;

    /**
     * Constructs an instance.
     *
     * @param unitCheck instance of running UnitCheck
     * @param testMethod test being checked
     * @param listeners listeners that will be notifies with summarized information
     * @param haltProperty property used for stopping JPF
     */
    public JPFJUnitListener(UnitChecker unitCheck, TestMethod testMethod,
            List<UnitCheckListener> listeners, StopExecutionProperty haltProperty) {
        this.unitCheck = unitCheck;
        this.listeners = listeners;
        this.testMethod = testMethod;
        this.haltProperty = haltProperty;
        this.runResult = new CheckRunResult();
    }

    // ================================= JPF ================================

    @Override
    public void propertyViolated(Search search) {
        log.trace("propertyViolated");
        Property prop = search.getLastError().getProperty();
        log.trace(prop.getClass().getName());

        // ignore our property
        if (prop instanceof StopExecutionProperty) {
            return;
        }
        processPropertyViolation(search);
        errorReported = true;
    }

    private void processPropertyViolation(Search search) {
        Error error = search.getLastError();
        log.trace("processPropertyViolation");
        PathHolder path = PathConvertor.convert(search.getVM().getPath(), unitCheck
                .getConfig());
        log.trace("processPropertyViolation1");
        Description desc = new Description(error.getDescription());
        String details = error.getDetails();

        PropertyFailure failure = new PropertyFailure(desc, details, SnapshotConvertor
                .convert(search.getVM()), path);
        runResult.addFailure(failure);
        runResult.checkFinished();

        for (UnitCheckListener l : listeners) {
            l.checkFinished(testMethod, failure);
        }
    }

    @Override
    public void searchConstraintHit(Search search) {
    }

    @Override
    public void searchFinished(Search search) {
    }

    @Override
    public void searchStarted(Search search) {
    }

    @Override
    public void stateAdvanced(Search search) {
    }

    @Override
    public void stateBacktracked(Search search) {
    }

    @Override
    public void stateProcessed(Search search) {
    }

    @Override
    public void stateRestored(Search search) {
    }

    @Override
    public void stateStored(Search search) {
    }

    // ================================= JUnit ================================

    @Override
    public void testAssumptionFailure(ExceptionFailure f) {
        log.trace(String.format("testAssumption=%s", f.getMessage()));
    }

    @Override
    public void testFailure(ExceptionFailure f) {
        log.trace(String.format("testFinished(FAILURE); message=%s, header=%s", f
                .getMessage(), f.getDescription().getAsString()));

        runResult.addFailure(f);
        runResult.checkFinished();
        errorReported = true;

        for (UnitCheckListener l : listeners) {
            l.checkFinished(testMethod, f);
        }

        if (unitCheck.getConfig().isHaltOnError()) {
            haltProperty.stop(StopExecutionProperty.EXCEPTION_CAUGHT_BY_JUNIT);
        }
    }

    @Override
    public void testFinished(Description description) {
        log.trace(String.format("testFinished=%s", description.getAsString()));

        if (!errorReported) {
            runResult.checkFinished();
            Success result = new Success();
            for (UnitCheckListener l : listeners) {
                l.checkFinished(testMethod, result);
            }
        }
        errorReported = false;
    }

    @Override
    public void testIgnored(Description description) {
        log.trace(String.format("testFinished=%s", description.getAsString()));
    }

    @Override
    public void testRunFinished(TestRunResult r) {
        log.trace(String.format(
                "testRunFinished: %s; failures=%d; ignored=%d; runCount=%d; time=%d ms",
                r.isSuccessful(), r.getFailureCount(), r.getIgnoreCount(),
                r.getRunCount(), r.getRunTime()));
    }

    @Override
    public void testRunStarted(Description description) {
        log.trace(String.format("testRunStarted"));
        for (UnitCheckListener l : listeners) {
            l.checkRunStarted(testMethod);
        }
    }

    @Override
    public void testStarted(Description description) {
        log.trace(String.format("testStarted"));
    }

    // =============================================================================

    /**
     * Called when checking of a method was finished.
     *
     * @param interrupted was the checking interrupted by a user?
     */
    public void finish(boolean interrupted) {
        for (UnitCheckListener l : listeners) {
            if (interrupted) {
                runResult.setInterrupted(true);
                l.checkRunInterrupted(testMethod, runResult);
            } else {
                l.checkRunFinished(testMethod, runResult);
            }
        }
    }

    // =============================================================================

}
